/**********************************************************
OpenCL Code for Matrix Multiplication.
Author: Petros Panayi,  PhD, EPL372 CS of UCY
Date: 07/04/2015
*** HowTo Run on euclid.cyi.ac.cy
From 103ws25.in.cs.ac.ucy copy the files to euclid:
scp * epl655u03@euclid.cyi.ac.cy:~/openclLab
And then Connect to euclide:
ssh epl655u03@euclid.cyi.ac.cy
module load CUDA/5.5.22
module load goolf/1.6.10
gcc -Wall -Werror -lOpenCL matrixmul_host.c -o matrixmul
or for DEBUG
gcc -Wall -Werror -DDEBUG -lOpenCL matrixmul_host.c -o matrixmul
salloc -N 1 --gres=gpu:2
./matrixmul
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
References:
https://www.khronos.org/opencl/
https://www.khronos.org/registry/cl/sdk/1.1/docs/man/xhtml/
https://www.khronos.org/registry/cl/specs/opencl-2.1.pdf
Devices: multi-core CPUs, GPUs, and other Accelerators 
Each Device contains one or more Compute Units, i.e. cores, SMs, etc...
Each Compute Unit contains one or more SIMD Processing Elements
OpenCL Context: Contains one or more devices
OpenCL memory objects are associated with a context, not a specific device
Memory transfers are associated with a command queue
OpenCL Program: contains one or more “kernels” and any supporting routines that run on a target device
OpenCL kernel: is the basic unit of code that can be executed on a target device

A OpenCL Platform --> N*Compute Devices --> N*Compute Units --> N*Processing Elements

OpenCL					CUDA
--------------------------------------
kernel					kernel
host program			host program
NDRange (index space)	grid
work item				thread
work group				block
global memory			global memory
constant memory			constant memory
local memory			shared memory
private memory			local memory
get_global_id(0);		blockIdx.x×blockDim.x+threadIdx.x
get_local_id(0);		blockIdx.x
get_global_size(0);		gridDim.x ×blockDim.x
get_local_size(0);		blockDim.x

*************************************************************/

////////////////////////////////////////////////////////////////////////////////

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <CL/cl.h>
#include <stdbool.h>
#include <assert.h>

#include "support.h"

////////////////////////////////////////////////////////////////////////////////
#define WA 1024
#define HA 1024
#define WB 1024

#define HB WA
#define WC WB
#define HC HA
////////////////////////////////////////////////////////////////////////////////

 
// Allocates a matrix with random float entries.
void randomMemInit(float* data, int size)
{
   int i;
   srand(time(NULL));
   for (i = 0; i < size; ++i)
   	data[i] = rand() / (float)RAND_MAX;
}
// Add all the elements of a Square Matrix and set values to 0 (Zero)
 void verify_and_Clean(float *C, int size){
	int i,j;
	float sum;
	for(i=0;i<size;i++)
		for(j=0;j<size;j++){
			sum += C[i*size+j];
			C[i*size+j]=0;
			}
	printf("Verification Sum: %.2f\n",sum);
}

// Serial MM of a Square Matrix for verification/Time Comparison reasons
void serialMultiply(float* A, float* B, float* C, int Width) 
{
	int i,j,k;
 // Matrix Multiplication Computation
 printf("Serial Matrix Multiplication [%dx%d]...\n",Width,Width);
  for (i = 0; i < Width; i++){
    for (j = 0; j < Width; j++){
      float sum = 0;
      for (k = 0; k < Width; k++){
        sum += A[i * Width + k] * B[k * Width + j];
      }
      C[i * Width + j] = sum;
    }
  }
  verify_and_Clean(C, Width);
 }



long LoadOpenCLKernel(char const* path, char **buf, bool add_nul)
{
    FILE  *fp;
    size_t fsz;
    long   off_end;
    int    rc;

    /* Open the file */
    fp = fopen(path, "r");
    if( NULL == fp ) {
        return -1L;
    }

    /* Seek to the end of the file */
    rc = fseek(fp, 0L, SEEK_END);
    if( 0 != rc ) {
        return -1L;
    }

    /* Byte offset to the end of the file (size) */
    if( 0 > (off_end = ftell(fp)) ) {
        return -1L;
    }
    fsz = (size_t)off_end;

    /* Allocate a buffer to hold the whole file */
    *buf = (char *) malloc( fsz+(int)add_nul );
    if( NULL == *buf ) {
        return -1L;
    }

    /* Rewind file pointer to start of file */
    rewind(fp);

    /* Slurp file into buffer */
    if( fsz != fread(*buf, 1, fsz, fp) ) {
        free(*buf);
        return -1L;
    }

    /* Close the file */
    if( EOF == fclose(fp) ) {
        free(*buf);
        return -1L;
    }

    if( add_nul ) {
        /* Make sure the buffer is NUL-terminated, just in case */
        *buf[fsz] = '\0';
    }

    /* Return the file size */
    return (long)fsz;
}

int main(int argc, char** argv)
{
   int err;                            // error code returned from api calls
   cl_int result = 0;
   cl_platform_id *platform;		   //
   cl_device_id device_id;             // compute device id 
   cl_context context;                 // compute context
   cl_command_queue commands;          // compute command queue
   cl_program program;                 // compute program
   cl_kernel kernel;                   // compute kernel

#ifdef DEBUG 
  cl_uint devices;
  cl_device_id *device;
  cl_uint max_compute_units, max_clock_frequency;
  char *info;
  size_t infosize;
  cl_uint ii, jj;
#endif
    // OpenCL device memory for matrices
   cl_mem d_A;
   cl_mem d_B;
   cl_mem d_C;

 
   //Allocate host memory for matrices A and B
   unsigned int size_A = WA * HA;
   unsigned int mem_size_A = sizeof(float) * size_A;
   float* h_A = (float*) malloc(mem_size_A);
 
   unsigned int size_B = WB * HB;
   unsigned int mem_size_B = sizeof(float) * size_B;
   float* h_B = (float*) malloc(mem_size_B);

  
 
   //Allocate host memory for the result C
   unsigned int size_C = WC * HC;
   unsigned int mem_size_C = sizeof(float) * size_C;
   float* h_C = (float*) malloc(mem_size_C);
  
   //Initialize host memory
   randomMemInit(h_A, size_A);
   randomMemInit(h_B, size_B);
#ifdef DEBUG
   startTime();
   serialMultiply(h_A, h_B, h_C, WA); 
   stopTime();elapsedTime(); 
#endif  
  
   
   printf("Initializing OpenCL device...\n"); 
   startTime();
   
   cl_uint dev_cnt = 0;
   
   clGetPlatformIDs(0, 0, &dev_cnt);
	
   cl_platform_id platform_ids[100];
   //char platform_name[100];
   //size_t platform_name_size;
   //char *platform_name;
   
   
   //The list of platforms available can be obtained using the following function.
   //https://www.khronos.org/registry/cl/specs/opencl-2.1.pdf#page=62
   cl_uint platforms;
   result = clGetPlatformIDs(0, 0, &platforms);
   assert(result == CL_SUCCESS);
   platform = (cl_platform_id *) malloc((size_t) sizeof(cl_platform_id) * platforms);
   //clGetPlatformIDs: Obtain the list of platforms available.
   result = clGetPlatformIDs(platforms, platform, 0);
   assert(result == CL_SUCCESS);
   printf("  %d platform(s)\n\n", platforms);

#ifdef DEBUG
// int i;
for(ii=0; ii<platforms; ii++) {
    printf("----------------------------------------\n"
	   "Querying platform %u\n", ii);
	// clGetPlatformInfo: Get specific information about the OpenCL platform.  
    result = clGetPlatformInfo(platform[ii], CL_PLATFORM_NAME, 0, 0, &infosize);
    info = (char *) malloc((size_t) sizeof(char) * infosize);
    result = clGetPlatformInfo(platform[ii], CL_PLATFORM_NAME, infosize, info, 0);
    assert(result == CL_SUCCESS);
    printf("  Name: \"%s\"\n", info);
    free(info);
    result = clGetPlatformInfo(platform[ii], CL_PLATFORM_VERSION, 0, 0, &infosize);
    assert(result == CL_SUCCESS);
    info = (char *) malloc((size_t) sizeof(char) * infosize);
    result = clGetPlatformInfo(platform[ii], CL_PLATFORM_VERSION, infosize, info, 0);
    assert(result == CL_SUCCESS);
    printf("  OpenCL version: \"%s\"\n", info);
    free(info);
    result = clGetPlatformInfo(platform[ii], CL_PLATFORM_PROFILE, 0, 0, &infosize);
    assert(result == CL_SUCCESS);
    info = (char *) malloc((size_t) sizeof(char) * infosize);
    result = clGetPlatformInfo(platform[ii], CL_PLATFORM_PROFILE, infosize, info, 0);
    assert(result == CL_SUCCESS);
    printf("  OpenCL profile: \"%s\"\n", info);
    free(info);
	//clGetDeviceIDs: Obtain the list of devices available on a platform.
    result = clGetDeviceIDs(platform[ii], CL_DEVICE_TYPE_ALL, 0, 0, &devices);
    assert(result == CL_SUCCESS);
    device = (cl_device_id *) malloc((size_t) sizeof(cl_device_id) * devices);
    result = clGetDeviceIDs(platform[0], CL_DEVICE_TYPE_ALL, devices, device, 0);
    assert(result == CL_SUCCESS);
    printf("  %d devices\n", devices);
  
    for(jj=0; jj<devices; jj++) {
      printf("  Querying device %u\n", jj);
	  //clGetDeviceInfo: Get information about an OpenCL device.
      result = clGetDeviceInfo(device[jj], CL_DEVICE_NAME, 0, 0, &infosize);
      assert(result == CL_SUCCESS);
      info = (char *) malloc((size_t) sizeof(char) * infosize);
      result = clGetDeviceInfo(device[jj], CL_DEVICE_NAME, infosize, info, 0);
      assert(result == CL_SUCCESS);
      printf("    Name: \"%s\"\n", info);
      free(info);
      
      result = clGetDeviceInfo(device[jj], CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(cl_uint), &max_compute_units, 0);
      assert(result == CL_SUCCESS);
      printf("    Compute units: %d\n", max_compute_units);
      
      result = clGetDeviceInfo(device[jj], CL_DEVICE_MAX_CLOCK_FREQUENCY, sizeof(cl_uint), &max_clock_frequency, 0);
      assert(result == CL_SUCCESS);
      printf("    Frequency: %d\n", max_clock_frequency);
    }
    free(device);
  }
  free(platform);
  printf("========================================\n");
#endif
  // Querying Platform Info: The list of platforms available can be obtained using the clGetPlatformIDs function.
  // clGetPlatformIDs: Obtain the list of platforms available. 
  // https://www.khronos.org/registry/cl/sdk/2.1/docs/man/xhtml/clGetPlatformIDs.html
   clGetPlatformIDs(dev_cnt, platform_ids, NULL);
  // Querying Devices: 
  // The list of devices available on a platform can be obtained using the clGetDeviceIDs function.
  // CL_DEVICE_TYPE_GPU: An OpenCL device that is a GPU. By this we mean that the device can also 
  //	be used to accelerate a 3D API such as OpenGL or DirectX.
   int gpu = 1;
   err = clGetDeviceIDs(platform_ids[0], gpu ? CL_DEVICE_TYPE_GPU : CL_DEVICE_TYPE_CPU, 1, &device_id, NULL);
   if (err != CL_SUCCESS)
   {
       printf("Error: Failed to create a device group!\n");
       return EXIT_FAILURE;
   }
   
   
   // Creates an OpenCL context. An OpenCL context is created with one or more devices. 
   // Contexts are used by the OpenCL runtime for managing objects such as command-queues, memory,
   // program and kernel objects and for executing kernels on one or more devices specified in the context.		
   // clCreateContext: Creates an OpenCL context.
   //https://www.khronos.org/registry/cl/sdk/2.1/docs/man/xhtml/clCreateContext.html
   //	num_devices[arg.2] is the number of devices specified in the devices argument.
   //	devices[arg.3] is a pointer to a list of unique devices returned by clGetDeviceIDs for a platform.
   context = clCreateContext(0, 1, &device_id, NULL, NULL, &err);
   if (!context)
   {
       printf("Error: Failed to create a compute context!\n");
       return EXIT_FAILURE;
   }
   // The command-queue can be used to queue a set of operations (referred to as commands) in order. 
   // Having multiple command-queues allows applications to queue multiple independent commands without
   // requiring synchronization.	
   //clCreateCommandQueue: Create a command commands
   //https://www.khronos.org/registry/cl/sdk/1.0/docs/man/xhtml/clCreateCommandQueue.html
   commands = clCreateCommandQueue(context, device_id, 0, &err);
   if (!commands)
   {
       printf("Error: Failed to create a command commands!\n");
       return EXIT_FAILURE;
   }

   // Create the compute program from the source file
   char *KernelSource;
   long lFileSize;

   // An Internal Function that loads the Kernel Code into &KernelSource
   lFileSize = LoadOpenCLKernel("matrixmul_kernel.cl", &KernelSource, false);
   if( lFileSize < 0L ) {
       perror("File read failed");
       return 1;
   }
	//clCreateProgramWithSource
	//https://www.khronos.org/registry/cl/sdk/2.1/docs/man/xhtml/clCreateProgramWithSource.html
	//Creates a program object for a context, and loads the source code
	//specified by the text strings in the strings array into the program object.
   program = clCreateProgramWithSource(context, 1, (const char **) & KernelSource, NULL, &err);
   if (!program)
   {
       printf("Error: Failed to create compute program!\n");
       return EXIT_FAILURE;
   }

   // clBuildProgram
   //https://www.khronos.org/registry/cl/sdk/2.1/docs/man/xhtml/clBuildProgram.html
   //Builds (compiles and links) a program executable from the program source or binary.
   err = clBuildProgram(program, 0, NULL, NULL, NULL, NULL);
   if (err != CL_SUCCESS)
   {
       size_t len;
       char buffer[2048];
       printf("Error: Failed to build program executable!\n");
	   //clGetProgramBuildInfo: returns build information for each device in the program object.
       clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, sizeof(buffer), buffer, &len);
       printf("%s\n", buffer);
       exit(1);
   }

   //clCreateKernel: Creates a kernal object.
   //https://www.khronos.org/registry/cl/sdk/2.1/docs/man/xhtml/clCreateKernel.html
   kernel = clCreateKernel(program, "matrixMul", &err);
   if (!kernel || err != CL_SUCCESS)
   {
       printf("Error: Failed to create compute kernel!\n");
       exit(1);
   }

   // Create the input and output buffer (arrays) in device memory for our calculation
   // CL_MEM_READ_WRITE: This flag specifies that the memory object will be read and written by a kernel.( This is the default.)
   // CL_MEM_COPY_HOST_PTR: Indicates that the application wants the OpenCL implementation to 
   //	allocate memory for the memory object and copy the data from memory referenced by host_ptr.
   //https://www.khronos.org/registry/cl/sdk/2.1/docs/man/xhtml/clCreateBuffer.html
   d_C = clCreateBuffer(context, CL_MEM_READ_WRITE, mem_size_A, NULL, &err);
   d_A = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, mem_size_A, h_A, &err);
   d_B = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, mem_size_B, h_B, &err);

   if (!d_A || !d_B || !d_C)
   {
       printf("Error: Failed to allocate device memory!\n");
       exit(1);
   }    
    
   printf("Running matrix multiplication for matrices A (%dx%d) and B (%dx%d) ...\n", WA,HA,WB,HB); 

   //Launch OpenCL kernel
   size_t localWorkSize[2], globalWorkSize[2];
 
   int wA = WA;
   int wC = WC;
   //clSetKernelArg: Used to set the argument value for a specific argument of a kernel.
   err = clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *)&d_C);
   err |= clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *)&d_A);
   err |= clSetKernelArg(kernel, 2, sizeof(cl_mem), (void *)&d_B);
   err |= clSetKernelArg(kernel, 3, sizeof(int), (void *)&wA);
   err |= clSetKernelArg(kernel, 4, sizeof(int), (void *)&wC);

   if (err != CL_SUCCESS)
   {
       printf("Error: Failed to set kernel arguments! %d\n", err);
       exit(1);
   }
   // global_work_size: points to an array of work_dim unsigned values that describe the number of
   //	global work-items in work_dim dimensions that will execute the kernel function. The total
   //	number of global work-items is computed as global_work_size[0] * … *
   //	global_work_size[work_dim – 1].
   
   // local_work_size: points to an array of work_dim unsigned values that describe the number of
   // 	work-items that make up a work-group
   localWorkSize[0] = 16;
   localWorkSize[1] = 16;
   globalWorkSize[0] = 1024;
   globalWorkSize[1] = 1024;
   //clEnqueueNDRangeKernel: Enqueues a command to execute a kernel on a device.
   // https://www.khronos.org/registry/cl/sdk/2.1/docs/man/xhtml/clEnqueueNDRangeKernel.html
   // In this example we have a 1024x1024 Matrix. The size of the 2D NDRange is 1024x1024
   // The number of Work-Items is 16x16=256. That is there are 4096 Work-Groups
   err = clEnqueueNDRangeKernel(commands, kernel, 2, NULL, globalWorkSize, localWorkSize, 0, NULL, NULL);

   if (err != CL_SUCCESS)
   {
       printf("Error: Failed to execute kernel! %d\n", err);
       exit(1);
   }
 
   //Retrieve result from device
   err = clEnqueueReadBuffer(commands, d_C, CL_TRUE, 0, mem_size_C, h_C, 0, NULL, NULL);

   if (err != CL_SUCCESS)
   {
       printf("Error: Failed to read output array! %d\n", err);
       exit(1);
   }
 
   //print out the results
/*
   printf("\n\nMatrix C (Results)\n");
   int i;
   for(i = 0; i < size_C; i++)
   {
      printf("%f ", h_C[i]);
      if(((i + 1) % WC) == 0)
      printf("\n");
   }
   printf("\n");
*/
#ifdef DEBUG
   verify_and_Clean(h_C, WA);
#endif
   printf("Matrix multiplication completed...\n"); 
  
   //Shutdown and cleanup
   free(h_A);
   free(h_B);
   free(h_C);
 
   clReleaseMemObject(d_A);
   clReleaseMemObject(d_C);
   clReleaseMemObject(d_B);

   clReleaseProgram(program);
   clReleaseKernel(kernel);
   clReleaseCommandQueue(commands);
   clReleaseContext(context);
   stopTime(); elapsedTime(); 
   return 0;
}

/***********************************************************************
[epl655u03@e01 openclLab]$  gcc -DDEBUG -Wall -Werror -lOpenCL matrixmul_host.c -o matrixmul
[epl655u03@e01 openclLab]$ ./matrixmul
Start Timer.
Serial Matrix Multiplication [1024x1024]...
Verification Sum: 267997536.00
Stop Timer. Elapsed Time: 9.64 Sec.
Initializing OpenCL device...
Start Timer.
  1 platform(s)
----------------------------------------
Querying platform 0
  Name: "NVIDIA CUDA"
  OpenCL version: "OpenCL 1.1 CUDA 6.5.14"
  OpenCL profile: "FULL_PROFILE"
  2 devices
  Querying device 0
    Name: "Tesla T10 Processor"
    Compute units: 30
    Frequency: 1296
  Querying device 1
    Name: "Tesla T10 Processor"
    Compute units: 30
    Frequency: 1296
========================================
Running matrix multiplication for matrices A (1024x1024) and B (1024x1024) ...
Verification Sum: 267997536.00
Matrix multiplication completed...
Stop Timer. Elapsed Time: 2.15 Sec.
********************************************************************************************/
